CREATE TABLE voters (
    public_token varchar NOT NULL PRIMARY KEY,
    fullname varchar NOT NULL,
    email varchar,
    admin boolean NOT NULL DEFAULT FALSE,
    email_sent boolean NOT NULL DEFAULT FALSE
);

CREATE TABLE questions (
    name varchar NOT NULL PRIMARY KEY,
    position integer NOT NULL,
    title varchar NOT NULL,
    sentence text NOT NULL,
    min_choices integer NOT NULL DEFAULT 0,
    max_choices integer DEFAULT NULL,
    vote_open boolean NOT NULL DEFAULT FALSE,
    vote_count integer NOT NULL DEFAULT 0,
    CONSTRAINT min_choices_positive CHECK (min_choices >= 0),
    CONSTRAINT nb_choices_range_valid CHECK (max_choices IS NULL OR min_choices <= max_choices)
);

CREATE TABLE answers (
    question_name varchar NOT NULL REFERENCES questions(name) ON DELETE CASCADE ON UPDATE CASCADE,
    name varchar NOT NULL,
    position integer NOT NULL,
    checked boolean NOT NULL DEFAULT FALSE,
    sentence text NOT NULL,
    vote_count integer NOT NULL DEFAULT 0,
    PRIMARY KEY(question_name, name)
);

CREATE TABLE votes (
    question_name varchar NOT NULL REFERENCES questions(name) ON DELETE CASCADE ON UPDATE CASCADE,
    voter_token varchar NOT NULL REFERENCES voters(public_token) ON DELETE CASCADE ON UPDATE CASCADE,
    vote_time timestamp NOT NULL DEFAULT NOW(),
    PRIMARY KEY(question_name, voter_token)
);

CREATE FUNCTION vote(questions.name%TYPE, voters.public_token%TYPE, varchar ARRAY)
RETURNS varchar
LANGUAGE plpgsql
AS $$
BEGIN

        -- Rollback if the vote is not open, or the vote is nonexistent
        IF (SELECT COUNT(1) FROM questions WHERE name=$1 AND vote_open) != 1
        THEN
            RETURN 'Vote is not open, or nonexistent';
        END IF;

        -- Rollback if the number of answers is not correct
        IF array_length($3, 1) < (SELECT min_choices FROM questions WHERE name=$1)
        THEN
            RETURN 'Too few answers selected in the vote';
        END IF;

        -- Rollback if the number of answers is not correct
        IF array_length($3, 1) > (SELECT max_choices FROM questions WHERE name=$1)
        THEN
            RETURN 'Too many answers selected in the vote';
        END IF;

        -- Rollback if the answers are not in the set of valid answers
        -- this also checks that the elements in the answers set are unique
        IF array_length($3, 1) != (SELECT COUNT(1) FROM answers WHERE question_name=$1 AND name=ANY($3))
        THEN
            RETURN 'There are duplicate or invalid answers in the vote';
        END IF;

        -- Rollback if the voter does not exist
        IF (SELECT COUNT(1) FROM voters WHERE public_token = $2) != 1
        THEN
            RETURN 'The voter is not found';
        END IF;

        -- Rollback if the vote already took place
        IF (SELECT COUNT(1) FROM votes WHERE question_name=$1  AND voter_token=$2) != 0
        THEN
            RETURN 'The vote has already taken place';
        END IF;

        -- Mark the vote so that it is not possible to vote twice
        INSERT INTO votes(question_name, voter_token) VALUES($1, $2);

        -- Add 1 to the total vote count for the question
        UPDATE questions
        SET vote_count = vote_count + 1
        WHERE name=$1;

        -- Add 1 to the vote count of every selected answers
        UPDATE answers
        SET vote_count = vote_count + 1
        WHERE question_name = $1
          AND name = ANY($3);

        RETURN NULL;
END;
$$;



INSERT INTO questions VALUES('moral',     1, 'rapport moral',              'Rapport moral pas trop chelou ?',     1, 1);
INSERT INTO questions VALUES('financier', 2, 'rapport financier',          'Rapport financier pas trop chelou ?', 1, 1);
INSERT INTO questions VALUES('ca',        3, 'constitution du nouveau CA', 'Qui qui ya dans le CA ?',      1, NULL);

INSERT INTO answers VALUES('moral',     'oui',   1, FALSE, 'Oui !');
INSERT INTO answers VALUES('moral',     'non',   2, FALSE, 'Non');
INSERT INTO answers VALUES('moral',     'blanc', 3, FALSE, 'Blanc');

INSERT INTO answers VALUES('financier', 'oui',   1, FALSE, 'Oui !');
INSERT INTO answers VALUES('financier', 'non',   2, FALSE, 'Non');
INSERT INTO answers VALUES('financier', 'blanc', 3, FALSE, 'Blanc');

INSERT INTO answers VALUES('ca',        'a',     1, TRUE, 'Alice');
INSERT INTO answers VALUES('ca',        'b',     2, TRUE, 'Bob');
INSERT INTO answers VALUES('ca',        'c',     3, TRUE, 'Charline');
INSERT INTO answers VALUES('ca',        'd',     4, TRUE, 'Dédé');

INSERT INTO voters VALUES('8082552deba234dfceb0fb4f019c516c476f3c2163287faa681e1c8eee0ab022', 'Robert Hacker', 'robert@example.org');
INSERT INTO voters VALUES('953d2c15c76c7d2ef4867669d406c713a5e20663d57d6aa20b8ae8fd48dda309', 'Dédé', 'dd@example.org');


UPDATE questions SET vote_open=TRUE WHERE name='moral';
SELECT vote('moral', '8082552deba234dfceb0fb4f019c516c476f3c2163287faa681e1c8eee0ab022', ARRAY ['oui']);
UPDATE questions SET vote_open=FALSE WHERE name='moral';

UPDATE questions SET vote_open = TRUE WHERE name='financier';
